from typing import Dict, Any, Optional, List, get_args

from loguru import logger

from pai_llm.conversation.algorithm.base import BaseAlgorithm
from pai_llm.conversation.exceptions import StorageError, ConversationNotFoundError, MessageNotFoundError
from pai_llm.conversation.models import Conversation, Message, ToolCallContent
from pai_llm.conversation.storage.base import BaseStorage
from pai_llm.exceptions import ValidationError
from pai_llm.types.chat.completion_role import CompletionRole

DEFAULT_USER_ID = "default_user"


class HistoryManager:

    def __init__(self, storage: BaseStorage, algorithm: Optional[BaseAlgorithm] = None):
        assert storage is not None, "Storage is required"
        self.storage = storage
        self.algorithm = algorithm
        logger.debug(f"HistoryManager initialized with storage: {storage}, algorithm: {algorithm}")

    def create_conversation(
            self,
            user_id: Optional[str] = DEFAULT_USER_ID,
            metadata: Optional[Dict[str, Any]] = None
    ) -> Conversation:
        try:
            if user_id is None:
                raise ValidationError("User ID is required")
            conversation = Conversation(
                user_id=user_id,
                metadata=metadata if metadata else {},
            )
            self.storage.save_conversation(conversation)
            logger.debug(f"Conversation created for user: {user_id}, with id: {conversation.id}")
            return self.storage.get_conversation(conversation.id)
        except Exception as e:
            logger.error(f"Failed to create conversation: {e}")
            raise StorageError(f"Failed to create conversation: {e}") from e

    def list_conversations(self, user_id: str, page_no: int = 1, page_size: int = 10) -> List[Conversation]:
        if user_id is None:
            raise ValidationError("User ID is required")
        if page_no < 1 or page_size < 1:
            raise ValidationError("Page number and page size must be greater than 0")
        try:
            return self.storage.list_conversations(user_id, page_no, page_size)
        except Exception as e:
            logger.error(f"Failed to list conversations: {e}")
            raise StorageError(f"Failed to list conversations: {e}") from e

    def get_conversation(self, conversation_id: str) -> Conversation:
        if conversation_id is None:
            raise ValidationError("Conversation ID is required")
        try:
            conversation = self.storage.get_conversation(conversation_id)
            if conversation is None:
                logger.error(f"Conversation with id {conversation_id} not found")
                raise ConversationNotFoundError(f"Conversation with id {conversation_id} not found")

            if self.algorithm is not None:
                conversation.messages = self.algorithm.get_message_window(conversation.messages)
            logger.debug(f"Retrieved Conversion {conversation} with {len(conversation.messages)} messages")
            return conversation
        except ConversationNotFoundError:
            raise
        except Exception as e:
            logger.error(f"Failed to get conversation: {e}")
            raise StorageError(f"Failed to get conversation: {e}") from e

    def rename_conversation(self, conversation_id: str, new_name: str) -> bool:
        try:
            return self.storage.rename_conversation(conversation_id, new_name)
        except Exception as e:
            logger.error(f"Failed to rename conversation: {e}")
            raise StorageError(f"Failed to rename conversation: {e}") from e

    def delete_conversation(self, conversation_id: str) -> bool:
        try:
            return self.storage.delete_conversation(conversation_id)
        except Exception as e:
            logger.error(f"Failed to delete conversation: {e}")
            raise StorageError(f"Failed to delete conversation: {e}") from e

    def search_conversations(
            self,
            user_id: str,
            query: Dict[str, Any]
    ) -> List[Conversation]:
        if user_id is None:
            raise ValidationError("User ID is required")
        try:
            return self.storage.search_conversations(user_id, query)
        except Exception as e:
            logger.error(f"Failed to search conversations: {e}")
            raise StorageError(f"Failed to search conversations: {e}") from e

    def add_message(
            self,
            conversation_id: str,
            content: str,
            role: CompletionRole,
            metadata: Dict[str, Any] = None,
            tool_calls: List[ToolCallContent] = None,
            tool_call_id: str = None,
            tool_call_result: str = None
    ) -> Message:
        if conversation_id is None:
            raise ValidationError("Conversation ID is required")
        if content is not None and not isinstance(content, str):
            raise ValidationError("Content must be a string or None")
        allow_roles = get_args(CompletionRole)
        if role not in allow_roles:
            raise ValidationError(f"Invalid role: {role}, must be one of {allow_roles}")

        try:
            conversation = self.storage.get_conversation(conversation_id)
            if conversation is None:
                raise ConversationNotFoundError(f"Conversation with id {conversation_id} not found")

            message = Message(
                content=content,
                role=role,
                metadata=metadata,
                tool_calls=tool_calls,
                tool_call_id=tool_call_id,
                tool_call_result=tool_call_result
            )
            conversation.messages.append(message)
            conversation.updated_at = message.created_at
            self.storage.save_conversation(conversation)
            logger.debug(f"Message added to conversation with id: {conversation_id}")
            return message
        except ConversationNotFoundError:
            raise
        except ValidationError:
            raise
        except Exception as e:
            logger.error(f"Failed to add message: {e}")
            raise StorageError(f"Failed to add message: {e}") from e

    def get_messages(self, conversation_id: str) -> List[Message]:
        if conversation_id is None:
            raise ValidationError("Conversation ID is required")
        try:
            conversation = self.storage.get_conversation(conversation_id)
            if conversation is None:
                raise ConversationNotFoundError(f"Conversation with id {conversation_id} not found")
            return conversation.messages
        except ConversationNotFoundError:
            raise
        except Exception as e:
            logger.error(f"Failed to get messages: {e}")
            raise StorageError(f"Failed to get messages: {e}") from e

    def get_message(self, message_id: str) -> Message:
        if message_id is None:
            raise ValidationError("Message ID is required")
        try:
            message = self.storage.get_message(message_id)
            if message is not None:
                return message
            raise MessageNotFoundError(f"Message with id {message_id} not found")
        except ConversationNotFoundError:
            raise
        except Exception as e:
            logger.error(f"Failed to get message: {e}")
            raise StorageError(f"Failed to get message: {e}") from e



from typing_extensions import Literal, TypeAlias

SystemRole = "system"
AssistantRole = "assistant"
UserRole = "user"
DeveloperRole = "developer"
ToolRole = "tool"
FunctionRole = "function"

CompletionRole: TypeAlias = Literal["system", "user", "assistant", "developer", "tool", "function"]



import os
import sqlite3
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from loguru import logger

from pai_llm.conversation.exceptions import DatabaseOperationError
from pai_llm.conversation.storage.sql_storage import (
    SQLSchema,
    SQLStorage,
    SQLiteError,
    SQLConnection,
    SQLLiteRow,
)


class SQLiteSchema(SQLSchema):
    def create_conversation_table(self) -> str:
        return """
            CREATE TABLE IF NOT EXISTS conversation (
                id TEXT PRIMARY KEY,
                name TEXT NOT NULL,
                user_id TEXT NOT NULL,
                metadata TEXT,
                created_at TIMESTAMP NOT NULL,
                updated_at TIMESTAMP NOT NULL
            )
        """

    def create_message_table(self) -> str:
        return """
            CREATE TABLE IF NOT EXISTS message (
                id TEXT PRIMARY KEY,
                conversation_id TEXT NOT NULL,
                user_id TEXT NOT NULL,
                role TEXT NOT NULL,
                content TEXT NOT NULL,
                metadata TEXT,
                token_count INTEGER,
                message_index INTEGER,
                created_at TIMESTAMP NOT NULL,
                FOREIGN KEY (conversation_id) REFERENCES CONVERSATION(id) ON DELETE CASCADE
            )
        """

    def insert_conversation(self) -> str:
        return """
            INSERT OR REPLACE INTO conversation (id, name, user_id, metadata, created_at, updated_at)
            VALUES (?, ?, ?, ?, ?, ?)
        """

    def get_conversation(self) -> str:
        return """
            SELECT * FROM conversation WHERE id = ?
        """

    def list_conversations(self) -> str:
        return """
            SELECT * FROM conversation WHERE user_id = ? ORDER BY updated_at DESC LIMIT ? OFFSET ?
        """

    def rename_conversation(self) -> str:
        return """
            UPDATE conversation SET name = ? WHERE id = ?
        """

    def delete_conversation(self) -> str:
        return """
           DELETE FROM conversation WHERE id = ?
        """

    def insert_message(self) -> str:
        return """
            INSERT INTO message (id, conversation_id, user_id, role, content, metadata, token_count, message_index, created_at)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
        """

    def get_message(self) -> str:
        return """
            SELECT * FROM message WHERE id = ?
        """

    def get_messages(self) -> str:
        return """
            SELECT * FROM message WHERE conversation_id = ? ORDER BY message_index ASC
        """

    def list_messages(self) -> str:
        return """
            SELECT * FROM message WHERE conversation_id = ? ORDER BY message_index ASC LIMIT ?
        """

    def delete_messages(self) -> str:
        return """
            DELETE FROM message WHERE conversation_id = ?
        """


class SQLiteStorage(SQLStorage):

    def __init__(self, db_file_path: str | Path, max_messages: Optional[int] = 100):
        assert db_file_path, "Database file path cannot be empty"
        assert max_messages is None or max_messages > 0, "Max messages must be greater than 0"
        super().__init__(SQLiteSchema(), max_messages)

        self.db_file_path = db_file_path or ":memory:"
        if self.db_file_path != ":memory:":
            os.makedirs(os.path.dirname(os.path.abspath(self.db_file_path)), exist_ok=True)
        logger.debug(f"Initializing SQLiteStorage with db_file_path: {db_file_path}")

        try:
            with self._open_connection() as conn:
                self._init_tables(conn)
        except SQLiteError as e:
            logger.error(f"SQLiteStorage initialization failed: {e}")
            raise DatabaseOperationError(f"SQLiteStorage initialization failed: {e}") from e

    @contextmanager
    def _open_connection(self):
        conn: SQLConnection | None = None
        try:
            conn = sqlite3.connect(self.db_file_path)
            conn.row_factory = SQLLiteRow
            conn.execute("PRAGMA foreign_keys = ON")
            yield conn
        except Exception as e:
            logger.error(f"Failed to connect to SQLite database: {e}")
            if conn:
                conn.rollback()
            raise DatabaseOperationError(f"Failed to connect to SQLite database: {e}") from e
        finally:
            if conn:
                conn.close()

    def _get_param_placeholder(self) -> str:
        return "?"

    def _start_transaction(self, conn: SQLConnection):
        conn.execute("BEGIN TRANSACTION")



from pai_llm.exceptions import PaiLLMError


class ConversationNotFoundError(PaiLLMError):
    pass


class MessageNotFoundError(PaiLLMError):
    pass


class SQLiteStorageError(PaiLLMError):
    pass


class DatabaseOperationError(SQLiteStorageError):
    pass


class StorageError(PaiLLMError):
    pass


class UnsupportedContentTypeError(PaiLLMError):
    pass



import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional

from pai_llm.types.chat.completion_role import CompletionRole


@dataclass
class MessageContent:
    pass


## Basic MessageContent ##
@dataclass
class TextContent(MessageContent):
    text: str


@dataclass
class ImageContent(MessageContent):
    url: str
    format: str = field(default_factory=lambda: "png")
    detail: Optional[str] = None


@dataclass
class AudioContent(MessageContent):
    url: str
    detail: Optional[str] = None


@dataclass
class VideoContent(MessageContent):
    url: str
    detail: Optional[str] = None


@dataclass
class DocumentContent(MessageContent):
    url: str
    format: str = field(default_factory=lambda: "pdf")
    detail: Optional[str] = None


@dataclass
class ToolCallContent(MessageContent):
    id: str
    type: str
    function: Dict[str, any] = None


## Structure MessageContent ##
@dataclass
class JsonContent(MessageContent):
    data: Dict[str, any] = field(default_factory=dict),
    schema: Optional[str] = None


@dataclass
class XmlContent(MessageContent):
    pass


@dataclass
class MarkdownContent(MessageContent):
    pass


@dataclass
class HtmlContent(MessageContent):
    pass


@dataclass
class CsvContent(MessageContent):
    pass


## Fixed and Interactive MessageContent ##
@dataclass
class MultimodalContent(MessageContent):
    pass


@dataclass
class InteractiveContent(MessageContent):
    pass


@dataclass
class StreamContent(MessageContent):
    pass


@dataclass
class ChartContent(MessageContent):
    pass


@dataclass
class MapContent(MessageContent):
    pass


## Specified Application MessageContent ##
@dataclass
class CodeContent(MessageContent):
    pass


@dataclass
class TranslationContent(MessageContent):
    pass


@dataclass
class EmbeddingContent(MessageContent):
    pass


@dataclass
class Message:
    role: CompletionRole
    content: str | MessageContent
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    tool_calls: Optional[List[ToolCallContent]] = None
    tool_call_id: Optional[str] = None
    tool_call_result: Optional[str] = None
    metadata: Dict[str, any] = field(default_factory=dict)
    token_count: Optional[int] = None
    created_at: datetime = field(default_factory=lambda: datetime.now())


@dataclass
class Conversation:
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = field(default_factory=lambda: "new conversation")
    user_id: str = field(default_factory=str)
    metadata: Dict[str, any] = field(default_factory=dict)
    messages: List[Message] = field(default_factory=list)
    created_at: datetime = field(default_factory=lambda: datetime.now())
    updated_at: datetime = field(default_factory=lambda: datetime.now())

    def add_message(self, message: Message):
        self.messages.append(message)
        self.updated_at = datetime.now()



import os
from enum import Enum
from typing import Union, cast

from .unified_model import UnifiedModel


class Model(UnifiedModel, Enum):

    def __new__(cls, value: Union["Model", str]) -> "Model":
        return cast("Model", UnifiedModel.__new__(cls, value))

    DEFAULT = os.getenv("DEFAULT_MODEL", "gpt-4o")

    GPT_3_5_TURBO = "gpt-3.5-turbo"
    GPT_4 = "gpt-4"
    GPT_4_TURBO = "gpt-4-turbo"
    GPT_4O = "gpt-4o"
    GPT_4O_MINI = "gpt-4o-mini"
    GPT_4_5_PREVIEW = "gpt-4.5-preview"
    O1 = "o1"
    O1_PREVIEW = "o1-preview"
    O1_MINI = "o1-mini"
    O3_MINI = "o3-mini"

    GLM_4 = "glm-4"
    GLM_4V = "glm-4v"
    GLM_4V_FLASH = "glm-4v-flash"
    GLM_4V_PLUS_0111 = "glm-4v-plus-0111"
    GLM_4_PLUS = "glm-4-plus"
    GLM_4_AIR = "glm-4-air"
    GLM_4_AIR_0111 = "glm-4-air-0111"
    GLM_4_AIRX = "glm-4-airx"
    GLM_4_LONG = "glm-4-long"
    GLM_4_FLASHX = "glm-4-flashx"
    GLM_4_FLASH = "glm-4-flash"
    GLM_ZERO_PREVIEW = "glm-zero-preview"
    GLM_3_TURBO = "glm-3-turbo"

    # Groq platform models
    GROQ_LLAMA_3_1_8B = "llama-3.1-8b-instant"
    GROQ_LLAMA_3_3_70B = "llama-3.3-70b-versatile"
    GROQ_LLAMA_3_3_70B_PREVIEW = "llama-3.3-70b-specdec"
    GROQ_LLAMA_3_8B = "llama3-8b-8192"
    GROQ_LLAMA_3_70B = "llama3-70b-8192"
    GROQ_MIXTRAL_8_7B = "mixtral-8x7b-32768"
    GROQ_GEMMA_2_9B_IT = "gemma2-9b-it"

    # TogetherAI platform models support tool calling
    TOGETHER_LLAMA_3_1_8B = "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo"
    TOGETHER_LLAMA_3_1_70B = "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo"
    TOGETHER_LLAMA_3_1_405B = "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo"
    TOGETHER_LLAMA_3_3_70B = "meta-llama/Llama-3.3-70B-Instruct-Turbo"
    TOGETHER_MIXTRAL_8_7B = "mistralai/Mixtral-8x7B-Instruct-v0.1"
    TOGETHER_MISTRAL_7B = "mistralai/Mistral-7B-Instruct-v0.1"

    # SambaNova Cloud platform models support tool calling
    SAMBA_LLAMA_3_1_8B = "Meta-Llama-3.1-8B-Instruct"
    SAMBA_LLAMA_3_1_70B = "Meta-Llama-3.1-70B-Instruct"
    SAMBA_LLAMA_3_1_405B = "Meta-Llama-3.1-405B-Instruct"

    # SGLang models support tool calling
    SGLANG_LLAMA_3_1_8B = "meta-llama/Meta-Llama-3.1-8B-Instruct"
    SGLANG_LLAMA_3_1_70B = "meta-llama/Meta-Llama-3.1-70B-Instruct"
    SGLANG_LLAMA_3_1_405B = "meta-llama/Meta-Llama-3.1-405B-Instruct"
    SGLANG_LLAMA_3_2_1B = "meta-llama/Llama-3.2-1B-Instruct"
    SGLANG_MIXTRAL_NEMO = "mistralai/Mistral-Nemo-Instruct-2407"
    SGLANG_MISTRAL_7B = "mistralai/Mistral-7B-Instruct-v0.3"
    SGLANG_QWEN_2_5_7B = "Qwen/Qwen2.5-7B-Instruct"
    SGLANG_QWEN_2_5_32B = "Qwen/Qwen2.5-32B-Instruct"
    SGLANG_QWEN_2_5_72B = "Qwen/Qwen2.5-72B-Instruct"

    STUB = "stub"

    # Legacy anthropic models
    # NOTE: anthropic legacy models only Claude 2.1 has system prompt support
    CLAUDE_2_1 = "claude-2.1"
    CLAUDE_2_0 = "claude-2.0"
    CLAUDE_INSTANT_1_2 = "claude-instant-1.2"

    # Claude3 models
    CLAUDE_3_OPUS = "claude-3-opus-latest"
    CLAUDE_3_SONNET = "claude-3-sonnet-20240229"
    CLAUDE_3_HAIKU = "claude-3-haiku-20240307"
    CLAUDE_3_5_SONNET = "claude-3-5-sonnet-latest"
    CLAUDE_3_5_HAIKU = "claude-3-5-haiku-latest"
    CLAUDE_3_7_SONNET = "claude-3-7-sonnet-latest"

    # Nvidia models
    NVIDIA_NEMOTRON_340B_INSTRUCT = "nvidia/nemotron-4-340b-instruct"
    NVIDIA_NEMOTRON_340B_REWARD = "nvidia/nemotron-4-340b-reward"
    NVIDIA_YI_LARGE = "01-ai/yi-large"
    NVIDIA_MISTRAL_LARGE = "mistralai/mistral-large"
    NVIDIA_MIXTRAL_8X7B = "mistralai/mixtral-8x7b-instruct"
    NVIDIA_LLAMA3_70B = "meta/llama3-70b"
    NVIDIA_LLAMA3_1_8B_INSTRUCT = "meta/llama-3.1-8b-instruct"
    NVIDIA_LLAMA3_1_70B_INSTRUCT = "meta/llama-3.1-70b-instruct"
    NVIDIA_LLAMA3_1_405B_INSTRUCT = "meta/llama-3.1-405b-instruct"
    NVIDIA_LLAMA3_2_1B_INSTRUCT = "meta/llama-3.2-1b-instruct"
    NVIDIA_LLAMA3_2_3B_INSTRUCT = "meta/llama-3.2-3b-instruct"
    NVIDIA_LLAMA3_3_70B_INSTRUCT = "meta/llama-3.3-70b-instruct"

    # Gemini models
    GEMINI_2_0_FLASH = "gemini-2.0-flash-exp"
    GEMINI_2_0_FLASH_THINKING = "gemini-2.0-flash-thinking-exp"
    GEMINI_2_0_PRO_EXP = "gemini-2.0-pro-exp-02-05"
    GEMINI_2_0_FLASH_LITE_PREVIEW = "gemini-2.0-flash-lite-preview-02-05"
    GEMINI_1_5_FLASH = "gemini-1.5-flash"
    GEMINI_1_5_PRO = "gemini-1.5-pro"

    # Mistral AI models
    MISTRAL_3B = "ministral-3b-latest"
    MISTRAL_7B = "open-mistral-7b"
    MISTRAL_8B = "ministral-8b-latest"
    MISTRAL_CODESTRAL = "codestral-latest"
    MISTRAL_CODESTRAL_MAMBA = "open-codestral-mamba"
    MISTRAL_LARGE = "mistral-large-latest"
    MISTRAL_MIXTRAL_8x7B = "open-mixtral-8x7b"
    MISTRAL_MIXTRAL_8x22B = "open-mixtral-8x22b"
    MISTRAL_NEMO = "open-mistral-nemo"
    MISTRAL_PIXTRAL_12B = "pixtral-12b-2409"

    # Reka models
    REKA_CORE = "reka-core"
    REKA_FLASH = "reka-flash"
    REKA_EDGE = "reka-edge"

    # Cohere models
    COHERE_COMMAND_R_PLUS = "command-r-plus"
    COHERE_COMMAND_R = "command-r"
    COHERE_COMMAND_LIGHT = "command-light"
    COHERE_COMMAND = "command"
    COHERE_COMMAND_NIGHTLY = "command-nightly"

    # Qwen models (Aliyun)
    QWEN_MAX = "qwen-max"
    QWEN_PLUS = "qwen-plus"
    QWEN_TURBO = "qwen-turbo"
    QWEN_LONG = "qwen-long"
    QWEN_VL_MAX = "qwen-vl-max"
    QWEN_VL_PLUS = "qwen-vl-plus"
    QWEN_MATH_PLUS = "qwen-math-plus"
    QWEN_MATH_TURBO = "qwen-math-turbo"
    QWEN_CODER_TURBO = "qwen-coder-turbo"
    QWEN_2_5_CODER_32B = "qwen2.5-coder-32b-instruct"
    QWEN_2_5_VL_72B = "qwen2.5-vl-72b-instruct"
    QWEN_2_5_72B = "qwen2.5-72b-instruct"
    QWEN_2_5_32B = "qwen2.5-32b-instruct"
    QWEN_2_5_14B = "qwen2.5-14b-instruct"
    QWEN_QWQ_32B = "qwq-32b-preview"
    QWEN_QVQ_72B = "qvq-72b-preview"
    QWEN_QWQ_PLUS = "qwq-plus"

    # Yi models (01-ai)
    YI_LIGHTNING = "yi-lightning"
    YI_LARGE = "yi-large"
    YI_MEDIUM = "yi-medium"
    YI_LARGE_TURBO = "yi-large-turbo"
    YI_VISION = "yi-vision"
    YI_MEDIUM_200K = "yi-medium-200k"
    YI_SPARK = "yi-spark"
    YI_LARGE_RAG = "yi-large-rag"
    YI_LARGE_FC = "yi-large-fc"

    # DeepSeek models
    DEEPSEEK_CHAT = "deepseek-chat"
    DEEPSEEK_REASONER = "deepseek-reasoner"
    # InternLM models
    INTERNLM3_LATEST = "internlm3-latest"
    INTERNLM3_8B_INSTRUCT = "internlm3-8b-instruct"
    INTERNLM2_5_LATEST = "internlm2.5-latest"
    INTERNLM2_PRO_CHAT = "internlm2-pro-chat"

    # Moonshot models
    MOONSHOT_V1_8K = "moonshot-v1-8k"
    MOONSHOT_V1_32K = "moonshot-v1-32k"
    MOONSHOT_V1_128K = "moonshot-v1-128k"

    # SiliconFlow models support tool calling
    SILICONFLOW_DEEPSEEK_V2_5 = "deepseek-ai/DeepSeek-V2.5"
    SILICONFLOW_DEEPSEEK_V3 = "deepseek-ai/DeepSeek-V3"
    SILICONFLOW_INTERN_LM2_5_20B_CHAT = "internlm/internlm2_5-20b-chat"
    SILICONFLOW_INTERN_LM2_5_7B_CHAT = "internlm/internlm2_5-7b-chat"
    SILICONFLOW_PRO_INTERN_LM2_5_7B_CHAT = "Pro/internlm/internlm2_5-7b-chat"
    SILICONFLOW_QWEN2_5_72B_INSTRUCT = "Qwen/Qwen2.5-72B-Instruct"
    SILICONFLOW_QWEN2_5_32B_INSTRUCT = "Qwen/Qwen2.5-32B-Instruct"
    SILICONFLOW_QWEN2_5_14B_INSTRUCT = "Qwen/Qwen2.5-14B-Instruct"
    SILICONFLOW_QWEN2_5_7B_INSTRUCT = "Qwen/Qwen2.5-7B-Instruct"
    SILICONFLOW_PRO_QWEN2_5_7B_INSTRUCT = "Pro/Qwen/Qwen2.5-7B-Instruct"
    SILICONFLOW_THUDM_GLM_4_9B_CHAT = "THUDM/glm-4-9b-chat"
    SILICONFLOW_PRO_THUDM_GLM_4_9B_CHAT = "Pro/THUDM/glm-4-9b-chat"

    # AIML models support tool calling
    AIML_MIXTRAL_8X7B = "mistralai/Mixtral-8x7B-Instruct-v0.1"
    AIML_MISTRAL_7B_INSTRUCT = "mistralai/Mistral-7B-Instruct-v0.1"

    def __str__(self):
        return self.value

    @property
    def is_openai(self) -> bool:
        r"""Returns whether this type of models is an OpenAI-released model."""
        return self in {
            Model.GPT_3_5_TURBO,
            Model.GPT_4,
            Model.GPT_4_TURBO,
            Model.GPT_4O,
            Model.GPT_4O_MINI,
            Model.O1,
            Model.O1_PREVIEW,
            Model.O1_MINI,
            Model.O3_MINI,
            Model.GPT_4_5_PREVIEW,
        }

    @property
    def is_azure_openai(self) -> bool:
        r"""Returns whether this type of models is an OpenAI-released model
        from Azure.
        """
        return self in {
            Model.GPT_3_5_TURBO,
            Model.GPT_4,
            Model.GPT_4_TURBO,
            Model.GPT_4O,
            Model.GPT_4O_MINI,
        }

    @property
    def is_zhipuai(self) -> bool:
        r"""Returns whether this type of models is an ZhipuAI model."""
        return self in {
            Model.GLM_3_TURBO,
            Model.GLM_4,
            Model.GLM_4V,
            Model.GLM_4V_FLASH,
            Model.GLM_4V_PLUS_0111,
            Model.GLM_4_PLUS,
            Model.GLM_4_AIR,
            Model.GLM_4_AIR_0111,
            Model.GLM_4_AIRX,
            Model.GLM_4_LONG,
            Model.GLM_4_FLASHX,
            Model.GLM_4_FLASH,
            Model.GLM_ZERO_PREVIEW,
        }

    @property
    def is_anthropic(self) -> bool:
        r"""Returns whether this type of models is Anthropic-released model.

        Returns:
            bool: Whether this type of models is anthropic.
        """
        return self in {
            Model.CLAUDE_INSTANT_1_2,
            Model.CLAUDE_2_0,
            Model.CLAUDE_2_1,
            Model.CLAUDE_3_OPUS,
            Model.CLAUDE_3_SONNET,
            Model.CLAUDE_3_HAIKU,
            Model.CLAUDE_3_5_SONNET,
            Model.CLAUDE_3_5_HAIKU,
            Model.CLAUDE_3_7_SONNET,
        }

    @property
    def is_groq(self) -> bool:
        r"""Returns whether this type of models is served by Groq."""
        return self in {
            Model.GROQ_LLAMA_3_1_8B,
            Model.GROQ_LLAMA_3_3_70B,
            Model.GROQ_LLAMA_3_3_70B_PREVIEW,
            Model.GROQ_LLAMA_3_8B,
            Model.GROQ_LLAMA_3_70B,
            Model.GROQ_MIXTRAL_8_7B,
            Model.GROQ_GEMMA_2_9B_IT,
        }

    @property
    def is_together(self) -> bool:
        r"""Returns whether this type of models is served by Together AI."""
        return self in {
            Model.TOGETHER_LLAMA_3_1_405B,
            Model.TOGETHER_LLAMA_3_1_70B,
            Model.TOGETHER_LLAMA_3_3_70B,
            Model.TOGETHER_LLAMA_3_3_70B,
            Model.TOGETHER_MISTRAL_7B,
            Model.TOGETHER_MIXTRAL_8_7B,
        }

    @property
    def is_sambanova(self) -> bool:
        r"""Returns whether this type of model is served by SambaNova AI."""
        return self in {
            Model.SAMBA_LLAMA_3_1_8B,
            Model.SAMBA_LLAMA_3_1_70B,
            Model.SAMBA_LLAMA_3_1_405B,
        }

    @property
    def is_mistral(self) -> bool:
        r"""Returns whether this type of models is served by Mistral."""
        return self in {
            Model.MISTRAL_LARGE,
            Model.MISTRAL_NEMO,
            Model.MISTRAL_CODESTRAL,
            Model.MISTRAL_7B,
            Model.MISTRAL_MIXTRAL_8x7B,
            Model.MISTRAL_MIXTRAL_8x22B,
            Model.MISTRAL_CODESTRAL_MAMBA,
            Model.MISTRAL_PIXTRAL_12B,
            Model.MISTRAL_8B,
            Model.MISTRAL_3B,
        }

    @property
    def is_nvidia(self) -> bool:
        r"""Returns whether this type of models is a NVIDIA model."""
        return self in {
            Model.NVIDIA_NEMOTRON_340B_INSTRUCT,
            Model.NVIDIA_NEMOTRON_340B_REWARD,
            Model.NVIDIA_YI_LARGE,
            Model.NVIDIA_MISTRAL_LARGE,
            Model.NVIDIA_LLAMA3_70B,
            Model.NVIDIA_MIXTRAL_8X7B,
            Model.NVIDIA_LLAMA3_1_8B_INSTRUCT,
            Model.NVIDIA_LLAMA3_1_70B_INSTRUCT,
            Model.NVIDIA_LLAMA3_1_405B_INSTRUCT,
            Model.NVIDIA_LLAMA3_2_1B_INSTRUCT,
            Model.NVIDIA_LLAMA3_2_3B_INSTRUCT,
            Model.NVIDIA_LLAMA3_3_70B_INSTRUCT,
        }

    @property
    def is_gemini(self) -> bool:
        r"""Returns whether this type of models is Gemini model.

        Returns:
            bool: Whether this type of models is gemini.
        """
        return self in {
            Model.GEMINI_2_0_FLASH,
            Model.GEMINI_1_5_FLASH,
            Model.GEMINI_1_5_PRO,
            Model.GEMINI_2_0_FLASH_THINKING,
            Model.GEMINI_2_0_PRO_EXP,
            Model.GEMINI_2_0_FLASH_LITE_PREVIEW,
        }

    @property
    def is_reka(self) -> bool:
        r"""Returns whether this type of models is Reka model.

        Returns:
            bool: Whether this type of models is Reka.
        """
        return self in {
            Model.REKA_CORE,
            Model.REKA_EDGE,
            Model.REKA_FLASH,
        }

    @property
    def is_cohere(self) -> bool:
        r"""Returns whether this type of models is a Cohere model.

        Returns:
            bool: Whether this type of models is Cohere.
        """
        return self in {
            Model.COHERE_COMMAND_R_PLUS,
            Model.COHERE_COMMAND_R,
            Model.COHERE_COMMAND_LIGHT,
            Model.COHERE_COMMAND,
            Model.COHERE_COMMAND_NIGHTLY,
        }

    @property
    def is_yi(self) -> bool:
        r"""Returns whether this type of models is Yi model.

        Returns:
            bool: Whether this type of models is Yi.
        """
        return self in {
            Model.YI_LIGHTNING,
            Model.YI_LARGE,
            Model.YI_MEDIUM,
            Model.YI_LARGE_TURBO,
            Model.YI_VISION,
            Model.YI_MEDIUM_200K,
            Model.YI_SPARK,
            Model.YI_LARGE_RAG,
            Model.YI_LARGE_FC,
        }

    @property
    def is_qwen(self) -> bool:
        return self in {
            Model.QWEN_MAX,
            Model.QWEN_PLUS,
            Model.QWEN_TURBO,
            Model.QWEN_LONG,
            Model.QWEN_VL_MAX,
            Model.QWEN_VL_PLUS,
            Model.QWEN_MATH_PLUS,
            Model.QWEN_MATH_TURBO,
            Model.QWEN_CODER_TURBO,
            Model.QWEN_2_5_CODER_32B,
            Model.QWEN_2_5_VL_72B,
            Model.QWEN_2_5_72B,
            Model.QWEN_2_5_32B,
            Model.QWEN_2_5_14B,
            Model.QWEN_QWQ_32B,
            Model.QWEN_QVQ_72B,
            Model.QWEN_QWQ_PLUS,
        }

    @property
    def is_deepseek(self) -> bool:
        return self in {
            Model.DEEPSEEK_CHAT,
            Model.DEEPSEEK_REASONER,
        }

    @property
    def is_internlm(self) -> bool:
        return self in {
            Model.INTERNLM3_LATEST,
            Model.INTERNLM3_8B_INSTRUCT,
            Model.INTERNLM2_5_LATEST,
            Model.INTERNLM2_PRO_CHAT,
        }

    @property
    def is_moonshot(self) -> bool:
        return self in {
            Model.MOONSHOT_V1_8K,
            Model.MOONSHOT_V1_32K,
            Model.MOONSHOT_V1_128K,
        }

    @property
    def is_sglang(self) -> bool:
        return self in {
            Model.SGLANG_LLAMA_3_1_8B,
            Model.SGLANG_LLAMA_3_1_70B,
            Model.SGLANG_LLAMA_3_1_405B,
            Model.SGLANG_LLAMA_3_2_1B,
            Model.SGLANG_MIXTRAL_NEMO,
            Model.SGLANG_MISTRAL_7B,
            Model.SGLANG_QWEN_2_5_7B,
            Model.SGLANG_QWEN_2_5_32B,
            Model.SGLANG_QWEN_2_5_72B,
        }

    @property
    def is_siliconflow(self) -> bool:
        return self in {
            Model.SILICONFLOW_DEEPSEEK_V2_5,
            Model.SILICONFLOW_DEEPSEEK_V3,
            Model.SILICONFLOW_INTERN_LM2_5_20B_CHAT,
            Model.SILICONFLOW_INTERN_LM2_5_7B_CHAT,
            Model.SILICONFLOW_PRO_INTERN_LM2_5_7B_CHAT,
            Model.SILICONFLOW_QWEN2_5_72B_INSTRUCT,
            Model.SILICONFLOW_QWEN2_5_32B_INSTRUCT,
            Model.SILICONFLOW_QWEN2_5_14B_INSTRUCT,
            Model.SILICONFLOW_QWEN2_5_7B_INSTRUCT,
            Model.SILICONFLOW_PRO_QWEN2_5_7B_INSTRUCT,
            Model.SILICONFLOW_THUDM_GLM_4_9B_CHAT,
            Model.SILICONFLOW_PRO_THUDM_GLM_4_9B_CHAT,
        }

    @property
    def is_aiml(self) -> bool:
        return self in {
            Model.AIML_MIXTRAL_8X7B,
            Model.AIML_MISTRAL_7B_INSTRUCT,
        }



import typing

from pai_llm.conversation.algorithm.base import BaseAlgorithm
from pai_llm.conversation.models import Conversation, Message


class FIFOAlgorithm(BaseAlgorithm):

    def __init__(self, max_messages: typing.Optional[int] = 100):
        assert max_messages is not None and max_messages > 0, "max_messages must be greater than 0"
        super().__init__(max_messages)

    def add_message(self, conversation: Conversation, new_message: Message) -> None:
        conversation.add_message(new_message)
        conversation.messages = self.get_message_window(conversation.messages)

    def get_message_window(self, messages: typing.List[Message]) -> typing.List[Message]:
        if len(messages) > self.max_messages:
            return messages[-self.max_messages:]
        return messages



import typing
from functools import wraps

from commons_lang import object_utils
from commons_lang.text import string_utils
from loguru import logger
from openai import OpenAI, AsyncOpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionMessage,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionUserMessageParam,
    ChatCompletionDeveloperMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionFunctionMessageParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionContentPartImageParam
)

from pai_llm.conversation.algorithm.base import BaseAlgorithm
from pai_llm.conversation.exceptions import UnsupportedContentTypeError
from pai_llm.conversation.history import HistoryManager
from pai_llm.conversation.models import (
    Message,
    MessageContent,
    TextContent,
    ImageContent,
    ToolCallContent
)
from pai_llm.conversation.storage.base import BaseStorage
from pai_llm.exceptions import ConfigurationError, ValidationError
from pai_llm.types.chat.completion_role import (
    CompletionRole,
    SystemRole,
    AssistantRole,
    UserRole,
    DeveloperRole,
    ToolRole,
    FunctionRole
)

T = typing.TypeVar("T", bound=OpenAI | AsyncOpenAI)


def _convert_to_message(
        openai_message: typing.Union[typing.Dict[str, typing.Any], ChatCompletionMessage]
) -> Message:
    try:
        if isinstance(openai_message, typing.Dict):
            role = object_utils.get(openai_message, "role")
            allow_roles = typing.get_args(CompletionRole)
            if role not in allow_roles:
                raise ValidationError(f"Invalid role: {role}, must be one of {allow_roles}")

            content = object_utils.get(openai_message, "content")

            # Handle tool calls
            tool_calls = None
            openai_tool_calls = object_utils.get(openai_message, "tool_calls")
            if object_utils.is_not_none(openai_tool_calls):
                tool_calls = [
                    ToolCallContent(
                        id=object_utils.get(openai_tool_call, "id"),
                        type=object_utils.get(openai_tool_call, "type"),
                        function=object_utils.get(openai_tool_call, "function"),
                    ) for openai_tool_call in openai_tool_calls
                ]

            # Handle tool calls result
            tool_call_id = object_utils.get(openai_message, "tool_call_id")
            tool_call_result: str | None = None
            if ToolRole == role and string_utils.is_not_blank(tool_call_id):
                tool_call_result = object_utils.get(openai_message, "content")

            # Handle multimodal content
            if isinstance(content, typing.List):
                processed_contents: typing.List[MessageContent] = []
                for item in content:
                    t = object_utils.get(item, "type")
                    if "text" == t:
                        text = object_utils.get(item, "text")
                        processed_contents.append(TextContent(text))
                    elif "image_url" == t:
                        image_url = object_utils.get(item, "image_url", {})
                        url = object_utils.get(image_url, "url")
                        detail = object_utils.get(image_url, "detail")
                        processed_contents.append(
                            ImageContent(
                                url=url,
                                detail=detail
                            )
                        )
                    else:
                        logger.warning(f"Unsupported content type: {t}")
                content = processed_contents

            return Message(
                role=role,
                content=content,
                tool_calls=tool_calls,
                tool_call_id=tool_call_id,
                tool_call_result=tool_call_result
            )
        else:
            role = object_utils.get(openai_message, "role")
            allow_roles = typing.get_args(CompletionRole)
            if role not in allow_roles:
                raise ValidationError(f"Invalid role: {role}, must be one of {allow_roles}")

            content = object_utils.get(openai_message, "content")

            # Handle tool calls
            tool_calls = None
            openai_tool_calls = object_utils.get(openai_message, "tool_calls")
            if object_utils.is_not_none(openai_tool_calls):
                tool_calls = [
                    ToolCallContent(
                        id=object_utils.get(openai_tool_call, "id"),
                        type=object_utils.get(openai_tool_call, "type"),
                        function=object_utils.get(openai_tool_call, "function"),
                    ) for openai_tool_call in openai_tool_calls
                ]

            # Handle tool calls result
            tool_call_id = object_utils.get(openai_message, "tool_call_id")
            tool_call_result: str | None = None
            if ToolRole == role and string_utils.is_not_blank(tool_call_id):
                tool_call_result = object_utils.get(openai_message, "content")

            return Message(
                role=typing.cast(CompletionRole, role),
                content=content,
                tool_calls=tool_calls,
                tool_call_id=tool_call_id,
                tool_call_result=tool_call_result
            )
    except Exception as e:
        if isinstance(e, ValidationError):
            raise
        logger.error(f"Failed to convert OpenAI message to message: {e}")
        raise ValidationError(f"Failed to convert OpenAI message to message: {e}") from e


def _convert_to_openai_messages(
        messages: typing.Sequence[Message]
) -> typing.List[ChatCompletionMessageParam]:
    openai_message: typing.List[ChatCompletionMessageParam] = []

    for message in messages:
        content = _convert_content_to_openai_format(content=message.content)
        role = message.role

        if SystemRole == role:
            openai_message.append(
                ChatCompletionSystemMessageParam(
                    role="system",
                    content=typing.cast(str, content)
                )
            )
        elif AssistantRole == role:
            message_param: typing.Dict[str, typing.Any] = {"role": AssistantRole}
            if object_utils.is_not_none(content):
                message_param["content"] = typing.cast(str, content)
            tool_calls = object_utils.get(message, "tool_calls")
            if object_utils.is_not_none(tool_calls):
                message_param["tool_calls"] = [
                    {
                        "id": tool_call.id,
                        "type": tool_call.type,
                        "function": tool_call.function
                    } for tool_call in tool_calls
                ]
            openai_message.append(
                typing.cast(ChatCompletionAssistantMessageParam, message_param)
            )
        elif UserRole == role:
            if isinstance(content, str):
                openai_message.append(
                    ChatCompletionUserMessageParam(
                        role="user",
                        content=typing.cast(str, content)
                    )
                )
            else:
                content_parts = typing.cast(
                    typing.List[
                        typing.Union[
                            ChatCompletionContentPartTextParam,
                            ChatCompletionContentPartImageParam
                        ]
                    ],
                    content
                )
                openai_message.append(
                    ChatCompletionUserMessageParam(
                        role="user",
                        content=content_parts
                    )
                )
        elif DeveloperRole == role:
            openai_message.append(
                ChatCompletionDeveloperMessageParam(
                    role="developer",
                    content=typing.cast(str, content)
                )
            )
        elif ToolRole == role:
            openai_message.append(
                ChatCompletionToolMessageParam(
                    role="tool",
                    content=typing.cast(str, content),
                    tool_call_id=message.tool_call_id or ""
                )
            )
        elif FunctionRole == role:
            openai_message.append(
                ChatCompletionFunctionMessageParam(
                    role="function",
                    content=typing.cast(str, content)
                    # TODO 考虑如何填充
                    # name=message.tool_call_id or ""
                )
            )

    return openai_message


def _convert_content_to_openai_format(
        content: typing.Union[str, typing.List[MessageContent], None]
) -> typing.Union[str, typing.List[typing.Dict[str, typing.Any]], None]:
    if content is None:
        return None
    if isinstance(content, str):
        return content

    openai_format: typing.List[typing.Dict[str, typing.Any]] = []
    for item in content:
        if isinstance(item, TextContent):
            openai_format.append({"type": "text", "text": item.text})
        elif isinstance(item, ImageContent):
            image_url: typing.Dict[str, typing.Any] = {"url": item.url}
            openai_format.append(
                {
                    "type": "image_url",
                    "image_url": image_url
                }
            )
        else:
            raise UnsupportedContentTypeError(f"Unsupported content type: {type(item)}")
    return openai_format


def with_history(
        storage: typing.Optional[BaseStorage] = None,
        algorithm: typing.Optional[BaseAlgorithm] = None,
        history_manager: typing.Optional[HistoryManager] = None
) -> typing.Callable[[T], T]:
    if not storage and not history_manager:
        raise ConfigurationError("Either storage or history_manager must be provided.")

    manager = history_manager or HistoryManager(
        storage=typing.cast(BaseStorage, storage),
        algorithm=algorithm
    )

    def decorator(client: T) -> T:
        nonlocal manager
        original_create = client.chat.completions.create

        def _prepare_messages(
                conversation_id: str,
                openai_messages: typing.Sequence[
                    typing.Union[
                        typing.Dict[str, typing.Any],
                        ChatCompletionMessage
                    ]
                ]
        ) -> typing.List[ChatCompletionMessageParam]:

            conversation = manager.get_conversation(conversation_id)
            converted_messages = [_convert_to_message(message) for message in openai_messages]

            if conversation is None:
                return _convert_to_openai_messages(converted_messages)

            system_message = next(
                (message for message in converted_messages if message.role == SystemRole),
                None
            )

            tool_responses = [
                message for message in converted_messages
                if message.role == ToolRole
            ]

            assistant_tools_pairs = []
            if tool_responses:
                for tool_response in tool_responses:
                    tool_call_id = tool_response.tool_call_id
                    if not tool_call_id:
                        continue
                    assistant_with_tool_call = None
                    for message in conversation.messages:
                        if AssistantRole == message.role and message.tool_call_id:
                            for tc in message.tool_calls:
                                if tc.id == tool_call_id:
                                    assistant_with_tool_call = message
                                    break
                            if assistant_with_tool_call:
                                break

                    if assistant_with_tool_call:
                        assistant_tools_pairs.append(
                            (assistant_with_tool_call, tool_response)
                        )
            else:
                for message in conversation.messages:
                    if system_message and SystemRole == message.role:
                        continue
                    converted_messages.append(message)

            # Prepare conversation messages
            conversation_messages: typing.List[Message] = []

            # Add system message to the beginning of the conversation messages
            if system_message:
                conversation_messages.insert(0, system_message)

            for message in converted_messages:
                # Skip system message if it was handled
                if (
                        system_message
                        and SystemRole == message.role
                        and message.content == system_message.content
                ):
                    continue

                # Skip tool responses that we're handling
                if ToolRole == message.role and message.tool_call_id:
                    if any(message.tool_call_id == pair[1].tool_call_id for pair in assistant_tools_pairs):
                        continue

                conversation_messages.append(message)

            for assistant_message, tool_message in assistant_tools_pairs:
                conversation_messages.append(assistant_message)
                conversation_messages.append(tool_message)

            return _convert_to_openai_messages(conversation_messages)

        def _prepare_content_for_storage(content: typing.Union[str, typing.List[MessageContent], None]) -> str:
            if content is None:
                return ""

            if isinstance(content, str):
                return content

            text_paris = []
            for item in content:
                if isinstance(item, TextContent):
                    text_paris.append(item.text)
                elif isinstance(item, ImageContent):
                    text_paris.append(f"[Image: {item.url}]")
                elif isinstance(item, ToolCallContent):
                    text_paris.append(f"[ToolCall: {item.type} - {item.id}]")
                else:
                    logger.warning(f"Unsupported content type: {type(item)}")

            return " ".join(text_paris) if text_paris else ""

        def chat_completions_create(
                conversation_id: str,
                openai_messages: typing.Sequence[
                    typing.Union[
                        typing.Dict[str, typing.Any],
                        ChatCompletionMessage
                    ]
                ],
                response: typing.Any,
        ) -> ChatCompletion:
            if not isinstance(response, ChatCompletion):
                raise TypeError(f"Expected ChatCompletion, got {type(response)}")

            for openai_message in openai_messages:
                converted_message = _convert_to_message(openai_message)
                content_for_storage = _prepare_content_for_storage(converted_message.content)

                manager.add_message(
                    conversation_id=conversation_id,
                    content=content_for_storage,
                    role=converted_message.role,
                    metadata={"type": "input"},
                    tool_calls=converted_message.tool_calls,
                    tool_call_id=converted_message.tool_call_id,
                    tool_call_result=converted_message.tool_call_result
                )

            for choice in response.choices:
                if isinstance(choice.message, ChatCompletionMessage):
                    converted_message = _convert_to_message(choice.message)
                    content_for_storage = _prepare_content_for_storage(converted_message.content)
                    manager.add_message(
                        conversation_id=conversation_id,
                        content=content_for_storage,
                        role=converted_message.role,
                        metadata={
                            "type": "output",
                            "model": response.model,
                            "usage": response.usage.model_dump() if response.usage else None,
                            "finish_reason": choice.finish_reason
                        },
                        tool_calls=converted_message.tool_calls,
                        tool_call_id=converted_message.tool_call_id,
                        tool_call_result=converted_message.tool_call_result
                    )

            return response

        @wraps(original_create)
        def sync_chat_completions_create(
                *args: typing.Any,
                conversation_id: typing.Optional[str],
                **kwargs: typing.Any
        ) -> ChatCompletion:
            if not conversation_id:
                conversation = manager.create_conversation()
                conversation_id = conversation.id

            openai_messages = kwargs.get("messages", [])
            prepared_messages = _prepare_messages(conversation_id, openai_messages)
            kwargs["messages"] = prepared_messages

            response = original_create(*args, **kwargs)
            return chat_completions_create(conversation_id, prepared_messages, response)

        @wraps(original_create)
        async def async_chat_completions_create(
                *args: typing.Any,
                conversation_id: typing.Optional[str],
                **kwargs: typing.Any
        ) -> ChatCompletion:
            if not conversation_id:
                conversation = manager.create_conversation()
                conversation_id = conversation.id

            openai_messages = kwargs.get("messages", [])
            prepared_messages = _prepare_messages(conversation_id, openai_messages)
            kwargs["messages"] = prepared_messages

            response = await original_create(*args, **kwargs)
            return chat_completions_create(conversation_id, prepared_messages, response)

        if isinstance(client, OpenAI):
            client.chat.completions.create = sync_chat_completions_create
        elif isinstance(client, AsyncOpenAI):
            client.chat.completions.create = async_chat_completions_create
        else:
            raise TypeError(f"Expected OpenAI or AsyncOpenAI, got {type(client)}")

        return client

    return decorator



import typing
from abc import ABC, abstractmethod

from pai_llm.conversation.models import Conversation, Message


class BaseAlgorithm(ABC):

    def __init__(self, max_messages: typing.Optional[int] = 100):
        self.max_messages = max_messages

    @abstractmethod
    def add_message(self, conversation: Conversation, new_message: Message) -> None:
        pass

    @abstractmethod
    def get_message_window(self, messages: typing.List[Message]) -> typing.List[Message]:
        pass



from abc import ABC, abstractmethod
from typing import List, Dict, Optional

from pai_llm.conversation.models import (
    Conversation,
    Message
)


class BaseStorage(ABC):

    def __init__(self, max_messages: Optional[int] = 100):
        self.max_messages = max_messages

    @abstractmethod
    def save_conversation(self, conversation: Conversation):
        pass

    @abstractmethod
    def get_conversation(
            self,
            conversation_id: str,
            messages_limit: Optional[int] = None
    ) -> Optional[Conversation]:
        pass

    @abstractmethod
    def list_conversations(
            self,
            user_id: str,
            page_no: int = 1,
            page_size: int = 10
    ) -> List[Conversation]:
        pass

    @abstractmethod
    def rename_conversation(self, conversation_id: str, new_name: str) -> bool:
        pass

    @abstractmethod
    def delete_conversation(self, conversation_id: str) -> bool:
        pass

    @abstractmethod
    def search_conversations(
            self,
            user_id: str,
            query: Dict[str, any]
    ) -> List[Conversation]:
        pass

    @abstractmethod
    def get_message(self, message_id: str) -> Message | None:
        pass



from threading import Lock
from typing import ClassVar, Union, cast, Dict


class UnifiedModel(str):
    _cache: ClassVar[Dict[str, "UnifiedModel"]] = {}
    _lock: ClassVar[Lock] = Lock()

    def __new__(cls, value: Union["UnifiedModel", str]) -> "UnifiedModel":
        with cls._lock:
            if value not in cls._cache:
                instance = super().__new__(cls, value)
                cls._cache[value] = cast(UnifiedModel, instance)
            else:
                instance = cls._cache[value]
            return instance



import json
import sqlite3
from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Dict, Optional, List, Tuple

import mysql
from commons_lang import object_utils
from loguru import logger
from mysql.connector.abstracts import MySQLConnectionAbstract
from mysql.connector.pooling import PooledMySQLConnection

from pai_llm.conversation.exceptions import DatabaseOperationError
from pai_llm.conversation.models import Message, Conversation
from pai_llm.conversation.storage.base import BaseStorage
from pai_llm.exceptions import DataConvertError

SQLiteConnection = sqlite3.Connection
MySQLConnection = PooledMySQLConnection | MySQLConnectionAbstract
SQLConnection = SQLiteConnection | MySQLConnection

SQLiteError = sqlite3.Error
MySQLError = mysql.connector.Error
SQLError = SQLiteError | MySQLError

SQLiteIntegrityError = sqlite3.IntegrityError
MySQLIntegrityError = mysql.connector.errors.IntegrityError
SQIntegrityError = SQLiteIntegrityError | MySQLIntegrityError

SQLLiteRow = sqlite3.Row
MySQLRow = Dict
SQLRow = SQLLiteRow | MySQLRow


class SQLSchema(ABC):

    @abstractmethod
    def create_conversation_table(self) -> str:
        pass

    @abstractmethod
    def create_message_table(self) -> str:
        pass

    @abstractmethod
    def insert_conversation(self) -> str:
        pass

    @abstractmethod
    def get_conversation(self) -> str:
        pass

    @abstractmethod
    def list_conversations(self) -> str:
        pass

    @abstractmethod
    def rename_conversation(self) -> str:
        pass

    @abstractmethod
    def delete_conversation(self) -> str:
        pass

    @abstractmethod
    def insert_message(self) -> str:
        pass

    @abstractmethod
    def get_message(self) -> str:
        pass

    @abstractmethod
    def get_messages(self) -> str:
        pass

    @abstractmethod
    def list_messages(self) -> str:
        pass

    @abstractmethod
    def delete_messages(self) -> str:
        pass


class SQLStorage(BaseStorage):

    def __init__(
            self,
            sql_scheme: SQLSchema,
            max_messages: Optional[int] = 100
    ):
        super().__init__(max_messages)
        self.sql_scheme = sql_scheme

    @abstractmethod
    @contextmanager
    def _open_connection(self):
        pass

    @abstractmethod
    def _get_param_placeholder(self) -> str:
        pass

    @abstractmethod
    def _start_transaction(self, conn: SQLConnection):
        pass

    def _init_tables(self, conn: SQLConnection):
        try:
            conn.execute(self.sql_scheme.create_conversation_table())
            conn.execute(self.sql_scheme.create_message_table())
            conn.commit()
            logger.debug(f"SQStorage tables initialization success")
        except SQLError as e:
            logger.error(f"SQStorage tables initialization failed: {e}")
            raise DatabaseOperationError(f"SQStorage tables initialization failed: {e}") from e

    @staticmethod
    def _serialize_metadata(metadata: Dict[str, Any]) -> str:
        try:
            return json.dumps(metadata)
        except Exception as e:
            logger.error(f"Failed to serialize metadata: {e}")
            raise DataConvertError(f"Failed to serialize metadata: {e}") from e

    @staticmethod
    def _deserialize_metadata(metadata_str: str) -> Dict[str, Any]:
        try:
            return json.loads(metadata_str)
        except Exception as e:
            logger.error(f"Failed to deserialize metadata: {e}")
            raise DataConvertError(f"Failed to deserialize metadata: {e}") from e

    @staticmethod
    def _convert_datetime(dt: str | datetime) -> datetime:
        try:
            if isinstance(dt, str):
                return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")
            else:
                return dt
        except Exception as e:
            logger.error(f"Failed to convert datetime: {e}")
            raise DataConvertError(f"Failed to convert datetime: {e}") from e

    def _conversation_to_row(self, conversation: Conversation) -> Tuple:
        assert conversation is not None, "Conversation cannot be None"
        assert conversation.id is not None, "Conversation id cannot be None"

        try:
            return (
                conversation.id,
                conversation.name,
                conversation.user_id,
                self._serialize_metadata(conversation.metadata),
                conversation.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                conversation.updated_at.strftime("%Y-%m-%d %H:%M:%S")
            )
        except Exception as e:
            logger.error(f"Failed to convert conversation to row: {e}")
            raise DataConvertError(f"Failed to convert conversation to row: {e}") from e

    def _row_to_conversation(self, row: SQLRow, messages: List[Message]) -> Conversation:
        try:
            return Conversation(
                id=row["id"],
                name=row["name"],
                user_id=row["user_id"],
                messages=messages,
                metadata=self._deserialize_metadata(row["metadata"]),
                created_at=self._convert_datetime(row["created_at"]),
                updated_at=self._convert_datetime(row["updated_at"]),
            )
        except Exception as e:
            logger.error(f"Failed to convert row to conversation: {e}")
            raise DataConvertError(f"Failed to convert row to conversation: {e}") from e

    def _message_to_row(self, message: Message, conversation_id: str, user_id: str, message_index: int) -> Tuple:
        assert message is not None, "Message cannot be None"
        assert conversation_id is not None, "Conversation id cannot be None"
        assert user_id is not None, "User id cannot be None"
        assert message_index >= 0, "MessageIndex must be greater than or equal to 0"

        try:
            return (
                message.id,
                conversation_id,
                user_id,
                message.role,
                message.content,
                self._serialize_metadata(message.metadata),
                message.token_count,
                message_index,
                message.created_at.strftime("%Y-%m-%d %H:%M:%S")
            )
        except Exception as e:
            logger.error(f"Failed to convert message to row: {e}")
            raise DataConvertError(f"Failed to convert message to row: {e}") from e

    def _row_to_message(self, row: SQLRow) -> Message:
        assert row is not None, "Row cannot be None"
        try:
            return Message(
                id=row["id"],
                role=row["role"],
                content=row["content"],
                metadata=self._deserialize_metadata(row["metadata"]),
                token_count=row["token_count"],
                created_at=self._convert_datetime(row["created_at"]),
            )
        except (KeyError, ValueError) as e:
            logger.error(f"Failed to convert row to message: {e}")
            raise DataConvertError(f"Failed to convert row to message: {e}") from e

    def save_conversation(self, conversation: Conversation):
        assert conversation is not None, "Conversation cannot be None"
        assert conversation.id is not None, "Conversation id cannot be None"

        try:
            with self._open_connection() as conn:
                self._start_transaction(conn)

                conversation_row = self._conversation_to_row(conversation)
                conn.execute(self.sql_scheme.insert_conversation(), conversation_row)
                conn.execute(self.sql_scheme.delete_messages(), (conversation.id,))

                messages = conversation.messages
                if object_utils.is_not_empty(messages) and (len(messages) > self.max_messages):
                    messages = messages[-self.max_messages:]

                for i, message in enumerate(messages):
                    message_row = self._message_to_row(message, conversation.id, conversation.user_id, i)
                    conn.execute(self.sql_scheme.insert_message(), message_row)

                conn.commit()
                logger.debug(f"SQLStorage saved conversation {conversation.id}")
        except SQIntegrityError as e:
            logger.error(f"Database integrity error while saving conversation {conversation.id}: {e}")
            raise DatabaseOperationError(
                f"Database integrity error while saving conversation {conversation.id}: {e}"
            ) from e
        except SQLiteError as e:
            logger.error(f"Database error while saving conversation {conversation.id}: {e}")
            raise DatabaseOperationError(
                f"Database error while saving conversation {conversation.id}: {e}"
            ) from e
        except ValueError as e:
            logger.error(f"Value error while saving conversation {conversation.id}: {e}")
            raise DatabaseOperationError(
                f"Value error while saving conversation {conversation.id}: {e}"
            ) from e

    def get_conversation(
            self,
            conversation_id: str,
            messages_limit: Optional[int] = None
    ) -> Optional[Conversation]:
        assert conversation_id is not None, "Conversation id cannot be None"
        assert messages_limit is None or messages_limit > 0, "Messages limit must be greater than 0"

        try:
            with self._open_connection() as conn:
                conversation_row = conn.execute(self.sql_scheme.get_conversation(), (conversation_id,)).fetchone()
                if not conversation_row:
                    logger.debug(f"SQLStorage got conversation {conversation_id} not found")
                    return None

                if messages_limit is not None:
                    message_rows = conn.execute(
                        self.sql_scheme.list_messages(),
                        (conversation_id, messages_limit)
                    ).fetchall()
                else:
                    message_rows = conn.execute(
                        self.sql_scheme.get_messages(),
                        (conversation_id,)
                    ).fetchall()

                messages = [self._row_to_message(row) for row in message_rows]
                return self._row_to_conversation(conversation_row, messages)
        except SQLiteError as e:
            logger.error(f"Database error while getting conversation {conversation_id}: {e}")
            raise DatabaseOperationError(
                f"Database error while getting conversation {conversation_id}: {e}"
            ) from e

    def list_conversations(self, user_id: str, page_no: int = 1, page_size: int = 10) -> List[Conversation]:
        assert page_no >= 1, "Page number must be greater than or equal to 1"
        assert page_size >= 1, "Page size must be greater than or equal to 1"

        conversations = []
        try:
            offset = (page_no - 1) * page_size
            limit = page_size
            with self._open_connection() as conn:
                rows = conn.execute(
                    self.sql_scheme.list_conversations(),
                    (user_id, limit, offset)
                ).fetchall()
                conversations = [self._row_to_conversation(row, []) for row in rows]
        except SQLiteError as e:
            logger.error(f"Database error while listing conversations: {e}")
            raise DatabaseOperationError(f"Database error while listing conversations: {e}") from e
        finally:
            return conversations

    def rename_conversation(self, conversation_id: str, new_name: str) -> bool:
        assert conversation_id is not None, "Conversation id cannot be None"
        assert new_name is not None, "New name cannot be None"

        ret = True
        try:
            with self._open_connection() as conn:
                conn.execute(self.sql_scheme.rename_conversation(), (new_name, conversation_id))
                conn.commit()
                logger.debug(f"SQLStorage renamed conversation {conversation_id} to {new_name}")
        except SQLiteError as e:
            logger.error(f"Database error while renaming conversation {conversation_id} to {new_name}: {e}")
            ret = False
            raise DatabaseOperationError(f"Database error while listing conversations: {e}") from e
        finally:
            return ret

    def delete_conversation(self, conversation_id: str) -> bool:
        assert conversation_id is not None, "Conversation id cannot be None"

        ret = True
        try:
            with self._open_connection() as conn:
                conn.execute(self.sql_scheme.delete_conversation(), (conversation_id,))
                conn.commit()
                logger.debug(f"SQLStorage deleted conversation {conversation_id}")
        except SQLiteError as e:
            logger.error(f"Database error while deleting conversation {conversation_id}: {e}")
            ret = False
            raise DatabaseOperationError(f"Database error while listing conversations: {e}") from e
        finally:
            return ret

    def search_conversations(self, user_id: str, query: Dict[str, any]) -> List[Conversation]:
        # TODO 支持语义搜索
        assert query is not None, "Query cannot be None"

        conversations = []
        conditions = []
        params = []
        try:
            conditions.append(f"user_id = {self._get_param_placeholder()}")
            params.append(user_id)

            if "name" in query:
                if not isinstance(query["name"], str):
                    raise ValueError("Name search criteria must be a string")
                conditions.append(f"name LIKE {self._get_param_placeholder()}")
                params.append(f"%{query['name']}%")

            if "metadata" in query:
                if not isinstance(query["metadata"], dict):
                    raise ValueError("Metadata search criteria must be a dictionary")
                for key, value in query["metadata"].items():
                    conditions.append(f"json_extract(metadata, '$.{key}') = {self._get_param_placeholder()}")
                    params.append(value)

            if "content" in query:
                if not isinstance(query["content"], str):
                    raise ValueError("Content search criteria must be a string")
                conditions.append(
                    f"""
                    id IN (
                        SELECT DISTINCT conversation_id
                        from message
                        WHERE content LIKE {self._get_param_placeholder()} and user_id = {self._get_param_placeholder()}
                    )
                    """
                )
                params.append(f"%{query['content']}%")
                params.append(user_id)
            if not conditions:
                return []

            sql = f"""
            SELECT * FROM conversation
            WHERE {" AND ".join(conditions)}
            ORDER BY updated_at DESC
            """

            with self._open_connection() as conn:
                rows = conn.execute(sql, params).fetchall()
                for row in rows:
                    conversations.append(self._row_to_conversation(row, []))
        except SQLiteError as e:
            logger.error(f"Database error while searching conversations: {e}")
            raise DatabaseOperationError(f"Database error while listing conversations: {e}") from e
        finally:
            return conversations

    def get_message(self, message_id: str) -> Message | None:
        assert message_id is not None, "Message id cannot be None"
        try:
            with self._open_connection() as conn:
                message_row = conn.execute(self.sql_scheme.get_message(), (message_id,)).fetchone()
                if message_row is None:
                    logger.debug(f"SQLStorage message {message_id} not found")
                    return None
                return self._row_to_message(message_row)
        except SQLiteError as e:
            logger.error(f"Database error while getting message {message_id}: {e}")
            raise DatabaseOperationError(f"Database error while getting message {message_id}: {e}") from e



from contextlib import contextmanager
from typing import Optional

from commons_lang import object_utils
from loguru import logger
from mysql.connector import pooling, errorcode
from mysql.connector.cursor import MySQLCursor

from pai_llm.conversation.exceptions import DatabaseOperationError
from pai_llm.conversation.storage.sql_storage import (
    SQLConnection,
    SQLSchema,
    SQLStorage,
    MySQLError,
)


class MySQLConfig:
    def __init__(
            self,
            host: str = "localhost",
            port: int = 3306,
            user: str = "root",
            password: str = "",
            database: str = None,
            charset: str = "utf8mb4",
            autocommit: bool = False,
            connect_timeout: int = 10,
            pool_size: int = 5
    ):
        """
        MySQL connection parameters configuration class

        Parameters:
            host: Host address
            port: Port number
            user: Username
            password: Password
            database: Database name
            charset: Character set
            autocommit: Whether to auto-commit transactions
            connect_timeout: Connection timeout in seconds
            pool_size: Connection pool size
        """
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.charset = charset
        self.autocommit = autocommit
        self.connect_timeout = connect_timeout
        self.pool_size = pool_size

    def get_connection_params(self):
        """Returns connection parameters as a dictionary"""
        params = {
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password,
            "charset": self.charset,
            "connect_timeout": self.connect_timeout
        }

        # Only add database when it's not None
        if self.database:
            params["database"] = self.database

        return params

    def get_pool_params(self):
        """Returns connection pool parameters as a dictionary"""
        params = self.get_connection_params()
        params["pool_size"] = self.pool_size
        return params


class MySQLSchema(SQLSchema):

    def create_conversation_table(self) -> str:
        return """
            CREATE TABLE IF NOT EXISTS conversation (
                id VARCHAR(255) PRIMARY KEY,
                name VARCHAR(255) NOT NULL,
                user_id VARCHAR(255) NOT NULL,
                metadata TEXT,
                created_at TIMESTAMP NOT NULL,
                updated_at TIMESTAMP NOT NULL
            );
        """

    def create_message_table(self) -> str:
        return """
            CREATE TABLE IF NOT EXISTS message (
                id VARCHAR(255) PRIMARY KEY,
                conversation_id VARCHAR(255) NOT NULL,
                user_id VARCHAR(255) NOT NULL,
                role VARCHAR(50) NOT NULL,
                content TEXT NOT NULL,
                metadata TEXT,
                token_count INT,
                message_index INT,
                created_at TIMESTAMP NOT NULL,
                FOREIGN KEY (conversation_id) REFERENCES conversation(id) ON DELETE CASCADE
            );
        """

    def insert_conversation(self) -> str:
        return """
            INSERT INTO conversation (id, name, user_id, metadata, created_at, updated_at)
            VALUES (%s, %s, %s, %s, %s, %s)
            ON DUPLICATE KEY 
            UPDATE 
                name = VALUES(name),
                user_id = VALUES(user_id),
                metadata = VALUES(metadata),
                created_at = VALUES(created_at),
                updated_at = VALUES(updated_at);
        """

    def get_conversation(self) -> str:
        return """
            SELECT * FROM conversation WHERE id = %s
        """

    def list_conversations(self) -> str:
        return """
            SELECT * FROM conversation WHERE user_id = %s ORDER BY updated_at DESC LIMIT %s OFFSET %s
        """

    def rename_conversation(self) -> str:
        return """
            UPDATE conversation SET name = %s WHERE id = %s
        """

    def delete_conversation(self) -> str:
        return """
           DELETE FROM conversation WHERE id = %s
       """

    def insert_message(self) -> str:
        return """
            INSERT INTO message (id, conversation_id, user_id, role, content, metadata, token_count, message_index, created_at)
            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)
        """

    def get_message(self) -> str:
        return """
            SELECT * FROM message WHERE id = %s
        """

    def get_messages(self) -> str:
        return """
            SELECT * FROM message WHERE conversation_id = %s ORDER BY message_index ASC
        """

    def list_messages(self) -> str:
        return """
            SELECT * FROM message WHERE conversation_id = %s ORDER BY message_index ASC LIMIT %s
        """

    def delete_messages(self) -> str:
        return """
            DELETE FROM message WHERE conversation_id = %s
        """


class MySQLStorage(SQLStorage):

    def __init__(
            self,
            mysql_config: MySQLConfig,
            init_tables: bool = True,
            max_messages: Optional[int] = 100
    ):
        assert mysql_config is not None, "MySQLConfig cannot be None"
        assert max_messages is None or max_messages > 0, "Max messages must be greater than 0"

        super().__init__(MySQLSchema(), max_messages)
        try:
            self.pool = pooling.MySQLConnectionPool(
                pool_name="pai_llm_pool",
                **mysql_config.get_pool_params()
            )
        except MySQLError as e:
            errno = object_utils.get(e, "errno")
            if errno == errorcode.ER_ACCESS_DENIED_ERROR:
                raise DatabaseOperationError("MySQL connection failed: Invalid username or password")
            elif errno == errorcode.ER_BAD_DB_ERROR:
                raise DatabaseOperationError("MySQL connection failed: Database does not exist")
            else:
                raise DatabaseOperationError(f"MySQL connection failed: {e}") from e
        except Exception as e:
            logger.error(f"Failed to create MySQL connection pool: {e}")
            raise DatabaseOperationError(f"Failed to create MySQL connection pool: {e}") from e

        if init_tables:
            try:
                with self._open_connection() as conn:
                    self._init_tables(conn)
            except MySQLError as e:
                logger.error(f"MySQLStorage initialization failed: {e}")
                raise DatabaseOperationError(f"MySQLStorage initialization failed: {e}") from e

    @staticmethod
    def _dict_factory(cursor, row):
        d = {}
        for idx, col in enumerate(cursor.description):
            d[col[0]] = row[idx]
        return d

    @contextmanager
    def _open_connection(self):
        conn: SQLConnection | None = None
        cursor: MySQLCursor | None = None

        def execute(*args, **kwargs) -> MySQLCursor:
            nonlocal cursor
            cursor.execute(*args, **kwargs)
            return cursor

        try:
            conn = self.pool.get_connection()
            cursor = conn.cursor(dictionary=True)
            cursor.execute("SET FOREIGN_KEY_CHECKS = 1;")
            conn.execute = execute
            yield conn
        except Exception as e:
            logger.error(f"Failed to get MySQL connection: {e}")
            if conn:
                conn.rollback()
            raise DatabaseOperationError(f"Failed to get MySQL connection: {e}") from e
        finally:
            if cursor:
                cursor.close()
            if conn:
                conn.close()

    def _get_param_placeholder(self) -> str:
        return "%s"

    def _start_transaction(self, conn: SQLConnection):
        conn.execute("START TRANSACTION")
